# coding: utf-8

__all__ = ['send_json_request']

import httpx
import json
import sys
import logging
import time
import eikon.Profile

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

from .tools import is_string_type
from .eikonError import EikonError

__internal_timer__ = 15000


def send_json_request(entity, payload, debug=False):
    """
    Returns the JSON response.
    This function can be used for advanced usage or early access to new features.

    Parameters
    ----------
    entity: string
        A string containing a service name

    payload: string
        A string containing a JSON request

    debug: boolean, optional
        When set to True, the json request and response are printed.
        Default: False

    Returns
    -------
    string
        The JSON response as a string

    Raises
    ------
        EikonError
            If daemon is disconnected

        httpx.ConnectTimeout
            If request times out

        Exception
            If request fails (HTTP code other than 200)

        EikonError
            If daemon is disconnected
    """
    profile = eikon.Profile.get_profile()
    if profile:
        logger = profile.logger
        logger.trace('entity: {}'.format(entity))
        logger.trace('payload: {}'.format(payload))

        if not is_string_type(entity):
            error_msg = 'entity must be a string identifying an UDF endpoint'
            logger.error(error_msg)
            raise ValueError(error_msg)
        try:
            if is_string_type(payload):
                data = json.loads(payload)
            elif type(payload) is dict:
                data = payload
            else:
                error_msg = 'payload must be a string or a dictionary'
                logger.error(error_msg)
                raise ValueError(error_msg)
        except JSONDecodeError as e:
            error_msg = 'payload must be json well formed.\n'
            error_msg += str(e)
            logger.error(error_msg)
            raise e

        network_error = False
        try:
            # build the request
            udf_request = {'Entity': {'E': entity, 'W': data} }
            logger.debug('Request:{}'.format(udf_request))
            response = profile.send_request(json=udf_request)
            try:
                logger.debug('HTTP Response code: {}'.format(response.status_code))
                logger.debug('HTTP Response: {}'.format(response.text))
            except UnicodeEncodeError as unicode_error:
                _response = json.dumps(response.text, ensure_ascii=False).encode('utf8')
                logger.debug(f'HTTP Response unicode: {_response}')

            if response.status_code == 200:
                result = {}
                try:
                    result = response.json()
                    logger.trace('Response size: {}'.format(sys.getsizeof(json.dumps(result))))
                except JSONDecodeError:
                    logger.error(f"Failed to decode response to json: {response.text}")

                # Manage specifically DataGrid async mode
                if entity.startswith('DataGrid') and entity.endswith('Async'):
                    ticket = _check_ticket_async(result)
                    while ticket:
                        ticket_request = {'Entity': {
                                             'E': entity,
                                             'W': {'requests': [{'ticket': ticket}]}
                                         }}
                        logger.debug('Send ticket request:{}'.format(ticket_request))
                        response = profile.send_request(json=ticket_request, timeout=(15.0, 15.0))

                        try:
                            logger.debug(f'HTTP Response: {response.status_code} - {response.text}')
                        except UnicodeEncodeError:
                            _response = json.dumps(response.text, ensure_ascii=False).encode('utf8')
                            logger.debug(f'HTTP Response unicode: {_response}')
                        result = response.json()
                        ticket = _check_ticket_async(result)

                _check_server_error(result)
                return result
            else:
                _raise_for_status(response)
        except httpx.TimeoutException as timeout_exception:
            error_msg = f'HTTP TimeoutException: {timeout_exception}.'
            logger.error(error_msg)
            raise EikonError(408, error_msg)
        except EikonError as eikon_err:
            logger.error(f"HTTP request failed: {type(eikon_err).__name__}-{eikon_err.message}")
            raise EikonError(eikon_err.code, eikon_err.message)
        except httpx.ConnectTimeout as connect_timeout:
            logger.debug(f'HTTP ConnectTimeout: {connect_timeout}')
            network_error = True
        except httpx.HTTPError as http_error:
            logger.debug(f'HTTP Error: {http_error}')
            network_error = True
        except httpx.DecodingError as decoding_error:
            logger.debug(f'HTTP DecodingError: {decoding_error}')
            network_error = True
        except Exception as e:
            message = getattr(e, 'message', None) or e
            logger.error(f"HTTP request failed: {type(e).__name__}-{message}")
        if network_error:
            error_msg = 'Eikon Proxy not running or cannot be reached. Please read the documentation on troubleshooting'
            logger.error(error_msg)
            raise EikonError(401, error_msg)


def _check_ticket_async(server_response):
    """
    Check server response.

    Check is the server response contains a ticket.

    :param server_response: request's response
    :type server_response: requests.Response
    :return: ticket value if response contains a ticket, None otherwise
    """
    logger = eikon.Profile.get_profile().logger
    # ticket response should contains only one key
    if len(server_response) == 1:
        for key, value in server_response.items():
            ticket = value[0]
            if ticket and ticket.get('estimatedDuration'):
                ticket_duration = int(ticket['estimatedDuration'])
                ticket_duration = min(ticket_duration, __internal_timer__)
                ticket_value = ticket['ticket']
                message = 'Receive ticket from {}, wait for {} second'.format(key, ticket_duration / 1000.0)
                if ticket_duration > 1000:
                    message = message + 's'
                logger.info(message)
                time.sleep(ticket_duration / 1000.0)
                return ticket_value
    return None


def _check_server_error(server_response):
    """
    Check server response.

    Check is the server response contains an HTPP error or a server error.

    :param server_response: request's response
    :type server_response: requests.Response
    :return: nothing

    :raises: Exception('HTTP error : <error_message>) if response contains HTTP response
              ex: '<500 Server error>'
          or Exception('Server error (<error code>) : <server_response>') if UDF returns an error
              ex: {u'ErrorCode': 500, u'ErrorMessage': u'Requested datapoint was not found: News_Headlines', u'Id': u''}

    """
    logger = eikon.Profile.get_profile().logger
    str_response = str(server_response)

    # check HTTP response (server response is an object that can contain ErrorCode attribute)
    if hasattr(server_response, 'ErrorCode'):
        logger.error(getattr(server_response, 'ErrorMessage'))
        raise httpx.HTTPError(server_response)

    # check HTTPError on proxy request
    if str_response.startswith('<') and str_response.endswith('>'):
        logger.error(str_response)
        raise httpx.HTTPError(server_response)

    # check UDF response (server response is JSON and it can contain ErrorCode + ErrorMessage keys)
    if 'ErrorCode' in server_response and 'ErrorMessage' in server_response:
        error_message = server_response['ErrorMessage']
        if len(error_message.split(',')) > 4:
            status, reason_phrase, version, content, headers = error_message.split(',')[:5]
        logger.error(error_message)
        raise EikonError(int(server_response['ErrorCode']), error_message)

    # check DataGrid response (server response is JSON and it can contain error + transactionId keys)
    if 'error' in server_response and 'transactionId' in server_response:
        error_message = '{} (transactionId:{}'.format(server_response['error'],server_response['transactionId'])
        logger.error(error_message)
        raise EikonError(400, error_message)


def _raise_for_status(response):
    """Raises stored :class:`HTTPError`, if one occurred."""

    error_msg = ''
    _reason = None
    if hasattr(response, "reason"):
        _reason = response.reason
    elif hasattr(response, "reason_phrase"):
        _reason = response.reason_phrase

    if isinstance(_reason, bytes):
        # We attempt to decode utf-8 first because some servers
        # choose to localize their reason strings. If the string
        # isn't utf-8, we fall back to iso-8859-1 for all other
        # encodings. (See PR #3538)
        try:
            reason = _reason('utf-8')
        except UnicodeDecodeError:
            reason = _reason('iso-8859-1')
    else:
        reason = _reason

    logger = eikon.Profile.get_profile().logger

    # Check if retry-after is in headers
    retry_after = response.headers.get('retry-after', '0')

    # if eikon.Profile.get_profile().get_log_level() < logging.INFO:
    #     rate_limit = response.headers.get('x-ratelimit-limit')
    #     rate_remaining = response.headers.get('x-ratelimit-remaining')
    #     volume_limit = response.headers.get('x-volumelimit-limit')
    #     volume_remaining = response.headers.get('x-volumelimit-remaining')
    #
    #     logger.trace('Headers: x_ratelimit_limit={} / x_ratelimit_remaining={} '.format(rate_limit, rate_remaining))
    #     logger.trace('         x_volumelimit_limit={} / x_volumelimit_remaining={}'.format(volume_limit, volume_remaining))
    #     logger.trace('         retry_after {}'.format(retry_after))

    if 400 <= response.status_code < 500:
        error_msg = u'Client Error: %s' % response.text
    elif 500 <= response.status_code < 600:
        error_msg = u'Server Error: %s' % response.text

    if retry_after != '0':
        error_msg += ' Wait for {} second{}'.format(retry_after, '.' if retry_after == '1' else 's.')

    if error_msg:
        logger.error(u'Error code {} | {}'.format(response.status_code, error_msg))
        raise EikonError(response.status_code, error_msg)



# coding: utf8

__all__ = ['Stream', 'StreamState']

import json
import logging
from threading import Lock
from enum import Enum, unique
from threading import Event


@unique
class StreamState(Enum):
    """
    Define the state of the Stream.
        Closed : The Stream is closed and ready to be opened.
        Pending : the Stream is in a pending state.  Upon success, the Stream will move into an open state, otherwise will be closed.
        Open : The Stream is opened.
    """
    Closed = 1
    Pending = 2
    Open = 3


class Stream(object):

    def __init__(self, session=None):
        from ..Profile import get_profile

        self.__stream_lock = Lock()
        self._stream_id = None
        self._session = None

        self._name = None
        self._service = None
        self._fields = []
        self._domain = None

        self._state = StreamState.Closed

        if session is None:
            self._session = get_profile()._get_desktop_session()
        else:
            self._session = session

        if self._session is None:
            raise AttributeError("Session is mandatory")

        self._response_event = Event()
        self._session._register_stream(self)

    def __del__(self):
        self._session._unregister_stream(self)
        pass

    @property
    def stream_id(self):
        return self._stream_id

    @property
    def name(self):
        return self._name

    @property
    def state(self):
        return self._state

    async def _wait_for_response(self):
        if self.state is StreamState.Open:
            return True

        try:
            self._session.log(1, "Lock on wait for a response on stream {}".format(self._stream_id))
            self._response_event.wait()
            self._session.log(1, "Response was received on stream {}.".format(self._stream_id))
        except Exception as e:
            self._session.log(logging.ERROR, f'Error occurred on stream {self._stream_id}: {e!r}') #.format(self._stream_id, str(e)))
        self._response_event.clear()

    #######################################
    #  methods to open and close session  #
    #######################################
    def open(self, with_updates=True):
        """
        Open synchronously the data stream
        """
        self._session._loop.run_until_complete(self.open_async(with_updates=with_updates))
        return self._state

    def close(self):
        """
        Close the data stream
        """
        if self._state is StreamState.Open:
            self._session.log(logging.DEBUG, 'Close Stream subscription {}'.format(self._stream_id))

            mp_req_json = {
                'ID': self._stream_id,
                'Type': 'Close'
            }
            self._session.log(1,
                              'Sent close subscription:\n{}'.format(json.dumps(mp_req_json,
                                                                    sort_keys=True,
                                                                    indent=2,
                                                                    separators=(',', ':'))))
            self._session._send(mp_req_json)
        self._state = StreamState.Closed
        # Unblock any wait for response
        self._response_event.set()
        self._response_event.clear()
        return self._state

    ################################################
    #  methods to open asynchronously item stream  #
    ################################################
    async def open_async(self, with_updates=True):
        """
        Open asynchronously the data stream
        """
        from eikon.streaming_session.session import Session
        from eikon.eikonError import EikonError
        if self._session is None:
            raise AttributeError("Session is mandatory")

        if self._session.get_open_state() is Session.State.Closed:
            raise EikonError(-1, "Session must be opened")

        if self._state in [StreamState.Open, StreamState.Pending]:
            self._session.log(logging.DEBUG, 'Try to reopen asynchronously Stream {}'.format(self._stream_id))
            return self._state

        self._state = StreamState.Pending

        # Wait for login successful before sending the request
        result = await self._session.wait_for_streaming()

        if result:
            self._session.log(logging.DEBUG, 'Open asynchronously Stream {}'.format(self._stream_id))
            if self._service is None:
                message = "Subscribe to {}".format(self._name, self._fields)
            else:
                message = "Subscribe to {}:{}".format(self._service, self._name)
            if self._fields:
                message = "{} on {}".format(message, self._fields)
            self._session.log(logging.DEBUG, message)

            mp_req_json = {
                'ID': self._stream_id,
                'Domain': self._domain,
                'Key': {
                    'Name': self._name
                },
                'Streaming': with_updates
            }
            if self._service:
                mp_req_json['Key']['Service'] = self._service
            if self._fields:
                mp_req_json['View'] = self._fields

            self._session._send(mp_req_json)
            self._session.log(1, 'Sent subscription request:\n{}'.format(
                json.dumps(mp_req_json, sort_keys=True, indent=2, separators=(',', ':'))))

            # wait for response message
            self._session.log(1, 'Wait for a response on Stream {}'.format(self._stream_id))
            await self._wait_for_response()

            if self.state is StreamState.Open:
                self._session.log(1, 'Just receive a response on Stream {} !!'.format(self._stream_id))
        else:
            self._state = StreamState.Closed
            self._session.log(1, 'Start streaming failed. Set stream {} as {}'.format(self._stream_id, self._state))


        return self._state

    ####################
    # Stream callbacks #
    ####################
    def _on_refresh(self, message):
        with self.__stream_lock:
            if self._state in [StreamState.Pending, StreamState.Open] :
                self._session.log(1, f'Receive message {message} on stream {self._stream_id} [{self._name}]')
                self._state = StreamState.Open
                self._session.log(1, 'Set stream {} as {}'.format(self._stream_id, self._state))
            if not self._response_event.is_set():
                self._response_event.set()

    def _on_update(self, update):
        with self.__stream_lock:
            if self._state is StreamState.Open:
                self._session.log(1, f'Stream {self._stream_id} [{self._name}] - Receive update {update}')

    def _on_status(self, status):
        with self.__stream_lock:
            self._session.log(1, f'Stream {self._stream_id} [{self._name}] - Receive status {status}')
            if not self._response_event.is_set():
                self._response_event.set()

    def _on_complete(self,):
        with self.__stream_lock:
            if self._state in [StreamState.Pending, StreamState.Open]:
                self._session.log(1, f'Stream {self._stream_id} [{self._name}] - Receive complete')

    async def _set_response_event(self):
        self._response_event.set()

    def _on_error(self, error):
        with self.__stream_lock:
            self._session.log(1, f'Stream {self._stream_id} [{self._name}] - Receive error {error}')

            if not self._response_event.is_set():
                self._response_event.set()

    def _on_stream_state(self, state):
        self._state = state



# coding: utf-8

__all__ = ['get_symbology']

import eikon.json_requests
from .tools import is_string_type, check_for_string_or_list_of_strings, check_for_string
import pandas as pd

Symbology_UDF_endpoint = 'SymbologySearch'
symbol_types = {"ric": "RIC", "isin": "ISIN", "cusip": "CUSIP", "sedol": "SEDOL",
                "ticker": "ticker", "lipperid": "lipperID", "imo": "IMO", "oapermid": "OAPermID"}


def get_symbology(symbol, from_symbol_type='RIC', to_symbol_type=None, raw_output=False, debug=False, bestMatch=True):
    """
    Returns a list of instrument names converted into another instrument code.
    For example: convert SEDOL instrument names to RIC names

    Parameters
    ----------
    symbol: string or list of strings
        Single instrument or list of instruments to convert.

    from_symbol_type: string
        Instrument code to convert from.
        Possible values: 'CUSIP', 'ISIN', 'SEDOL', 'RIC', 'ticker', 'lipperID', 'IMO'
        Default: 'RIC'

    to_symbol_type: string or list
        Instrument code to convert to.
        Possible values: 'CUSIP', 'ISIN', 'SEDOL', 'RIC', 'ticker', 'lipperID', 'IMO', 'OAPermID'
        Default: None  (means all symbol types are requested)

    raw_output: boolean, optional
        Set this parameter to True to get the data in json format
        if set to False, the function will return a data frame
        Default: False

    debug: boolean, optional
        When set to True, the json request and response are printed.
        Default: False

    bestMatch: boolean, optional
        When set to True, only primary symbol is requested.
        When set to false, all symbols are requested
        Default: True

    Returns
    -------
        If raw_output is set to True, the data will be returned in the json format.
        If raw_output is False (default value) the data will be returned as a pandas.DataFrame

        pandas.DataFrame content:
            - columns : Symbol types
            - rows : Symbol requested
            - cells : the symbols (None if not found)
            - symbol : The requested symbol

    Raises
    ------
        Exception
            If request fails or if server returns an error

        ValueError
            If a parameter type or value is wrong

    Examples
    --------
    >>> import eikon as ek
    >>> ek.set_app_key('set your app key here')
    >>> ISIN_codes = ek.get_symbology(["MSFT.O", "GOOG.O", "IBM.N"], from_symbol_type="RIC", to_symbol_type="ISIN")
    >>> ISIN_codes
                    ISIN
    MSFT.O  US5949181045
    GOOG.O  US02079K1079
    IBM.N   US4592001014
    """
    logger = eikon.Profile.get_profile().logger

    # check if symbol type is string or list of strings
    check_for_string_or_list_of_strings(symbol, 'symbol')
    if is_string_type(symbol): symbol = [symbol]

    # check if from_symbol type is string
    check_for_string(from_symbol_type, 'from_symbol_type')
    try:
        from_symbol_type = symbol_types[from_symbol_type.lower()]
    except:
        error_msg = 'from_symbol_type "' + from_symbol_type + '" should be in ' + [symbol_types[key] for key in
                                                                                   symbol_types].__str__()
        logger.error(error_msg)
        raise ValueError(error_msg)

    # if from_symbol_type is RIC, apply rics = [ric.upper() if ric.islower() else ric for ric in rics ] transformation
    if from_symbol_type == 'RIC':
        symbol = [ric.upper() if ric.islower() else ric for ric in symbol]

    # to_symbol_type to None means request all symbol types
    if to_symbol_type is not None:
        # otherwise check if to_symbol type is string or list of strings
        check_for_string_or_list_of_strings(to_symbol_type, 'to_symbol_type')
        if is_string_type(to_symbol_type):
            to_symbol_type = [to_symbol_type.strip()]
        try:
            to_symbol_type = [symbol_types[_.lower()] for _ in to_symbol_type]
        except:
            error_msg = 'All items in the parameter to_symbol should be in ' + [symbol_types[key] for key in
                                                                                symbol_types].__str__()
            logger.error(error_msg)
            raise ValueError(error_msg)

    payload = {'symbols': symbol, 'from': from_symbol_type, 'to': to_symbol_type, 'bestMatchOnly': bestMatch}
    result = eikon.json_requests.send_json_request(Symbology_UDF_endpoint, payload, debug=debug)

    if raw_output:
        return result
    else:
        if bestMatch:
            results_dict = dict([(_['symbol'], _['bestMatch']) for _ in result['mappedSymbols']])
        else:
            results_dict = dict([(_['symbol'], _) for _ in result['mappedSymbols']])
        if len(results_dict):
            df_result = pd.DataFrame(results_dict).transpose()
            if not df_result.empty:
                df_result = df_result.convert_dtypes()
        else:
            df_result = pd.DataFrame([])


        return df_result



from unittest import TestCase
from mock import patch, Mock, MagicMock
import sys
import json
import httpx
import logging
from eikon import EikonError


class Mock_Profile(object):
    session = None

    def __init__(self, application_key):
        self.app_key = application_key
        self.port = 3000
        self.url = 'http://localhost:{0}/api/v1/data'.format(self.port)
        self.session = httpx.AsyncClient()
        self.logger = logging.getLogger('pyeikon')
        setattr(self.logger, 'trace', lambda *args: self.logger.log(5, *args))

    def set_app_key(self, app_key):
        self.app_key = app_key

    def get_app_key(self):
        return self.app_key

    def get_url(self):
        return self.url

    def get_session(self):
        return self.session

    def get_timeout(self):
        return 60

    def get_log_level(self):
        return 100


class Mock_Response(object):
    def __init__(self, status_code, response):
        self._status_code = status_code
        self._json = response

    @property
    def status_code(self):
        return self._status_code

    def json(self, **kwargs):
        return str(self._json)

    def text(self):
        return str(self.json())


class TestSend_json_request(TestCase):
    @patch('eikon.Profile.get_profile')
    def test_raise_AttributeError_if_entity_not_string(self, mock_get_profile):
        mock_get_profile.return_value = Mock()
        from eikon.json_requests import send_json_request
        self.assertRaises(ValueError, send_json_request, 1234, {})
        self.assertRaises(ValueError, send_json_request, True, {})
        self.assertRaises(ValueError, send_json_request, {'entity': 'test'}, {})
        self.assertRaises(ValueError, send_json_request, ['entity'], {})

    @patch('eikon.Profile.get_profile')
    def test_raise_AttributeError_if_payload_not_dict(self, mock_get_profile):
        mock_get_profile.return_value = Mock()
        from eikon.json_requests import send_json_request
        self.assertRaises(ValueError, send_json_request, 'entity', 'payload')
        self.assertRaises(ValueError, send_json_request, 'entity', True)
        self.assertRaises(ValueError, send_json_request, 'entity', ['payload'])
        self.assertRaises(ValueError, send_json_request, 'entity', 12345)

    @patch('eikon.Profile.get_profile')
    def test_raise_payload_jsondecode(self, mock_get_profile):
        mock_get_profile.return_value = Mock()
        from eikon.json_requests import send_json_request
        from json import JSONDecodeError
        self.assertRaises(JSONDecodeError, send_json_request, 'entity', 'x')

    @patch('eikon.Profile.get_profile')
    def test_enter_on_Exception(self, mock_get_profile):
        mock_profile = Mock_Profile('12345')
        setattr(mock_profile, 'send_request', MagicMock(
            side_effect=AttributeError
        ))
        mock_get_profile.return_value = mock_profile

        from eikon.json_requests import send_json_request
        self.assertIsNone(send_json_request('entity', {}))
        mock_profile.send_request.assert_called_once()

    @patch('eikon.Profile.get_profile')
    def test_raise_EikonError_on_ConnectionError(self, mock_get_profile):
        mock_profile = Mock_Profile('12345')
        setattr(mock_profile, 'send_request', MagicMock(
            side_effect=httpx.ConnectTimeout('ConnectTimeout', request=MagicMock())
        ))
        mock_get_profile.return_value = mock_profile

        from eikon.json_requests import send_json_request
        with self.assertRaises(EikonError) as ex_info:
            send_json_request('entity', {})
        self.assertEqual(ex_info.exception.code, 408)
        mock_profile.send_request.assert_called_once()

    @patch('eikon.Profile.get_profile')
    def test_raise_EikonError_on_HTTPError(self, mock_get_profile):
        mock_profile = Mock_Profile('12345')
        setattr(mock_profile, 'send_request', MagicMock(
            side_effect=httpx.HTTPError('HTTPError', request=MagicMock())
        ))
        mock_get_profile.return_value = mock_profile

        from eikon.json_requests import send_json_request
        with self.assertRaises(EikonError) as ex_info:
            send_json_request('entity', {})
        self.assertEqual(ex_info.exception.code, 401)
        mock_profile.send_request.assert_called_once()

    @patch('eikon.Profile.get_profile')
    def test_raise_TimeoutError_on_ConnectionTimeout(self, mock_get_profile):
        mock_profile = Mock_Profile('12345')
        setattr(mock_profile, 'send_request', MagicMock(
            side_effect=httpx.TimeoutException('TimeoutException', request=MagicMock())
        ))
        mock_get_profile.return_value = mock_profile

        from eikon.json_requests import send_json_request
        with self.assertRaises(EikonError) as ex_info:
            send_json_request('entity', {}, debug=True)
        self.assertEqual(ex_info.exception.code, 408)
        mock_profile.send_request.assert_called_once()

    @patch('eikon.Profile.get_profile')
    def test_raise_EikonError_on_DecodingError(self, mock_get_profile):
        mock_profile = Mock_Profile('12345')
        from json import JSONDecodeError
        setattr(mock_profile, 'send_request', MagicMock(
            side_effect=httpx.DecodingError('DecodingError', request=MagicMock())
        ))
        mock_get_profile.return_value = mock_profile

        from eikon.json_requests import send_json_request
        with self.assertRaises(EikonError) as ex_info:
            send_json_request('entity', {}, debug=True)
        self.assertEqual(ex_info.exception.code, 401)
        mock_profile.send_request.assert_called_once()

    @patch('eikon.Profile.get_profile')
    def test_raise_exception_when_post_status_code_error(self, mock_get_profile):
        mock_profile = Mock_Profile('12345')
        setattr(mock_profile, 'send_request', MagicMock(
            return_value=Mock_Response(400, 'error')
        ))
        mock_profile.send_request.return_value.headers = {}
        mock_get_profile.return_value = mock_profile

        from eikon.json_requests import send_json_request
        with self.assertRaises(EikonError) as ex_info:
            send_json_request('entity', {})
        self.assertEqual(ex_info.exception.code, 400)

    @patch('eikon.Profile.get_profile')
    @patch('eikon.Profile.get_url')
    def test_result_when_post_status_code_200(self, mock_get_url, mock_get_profile):
        mock_profile = Mock_Profile('12345')
        setattr(mock_profile, 'send_request', MagicMock(
            return_value=Mock_Response(200, {"key1": "value1"})
        ))
        mock_get_profile.return_value = mock_profile
        mock_get_url.return_value = 'http://localhost:3000/api/v1/data'

        from eikon.json_requests import send_json_request
        self.assertEqual("{'key1': 'value1'}", send_json_request('entity', {}))

    @patch('eikon.Profile.get_profile')
    def test_result_when_post_datagrib_async_code_200(self, mock_get_profile):
        mock_profile = Mock_Profile('12345')
        setattr(mock_profile, 'send_request', MagicMock(
            return_value=Mock_Response(200, {"key1": "value1"})
        ))
        mock_get_profile.return_value = mock_profile
        from eikon.json_requests import send_json_request
        from eikon.data_grid import DataGridAsync_UDF_endpoint
        self.assertEqual("{'key1': 'value1'}", send_json_request(DataGridAsync_UDF_endpoint, {}))

    @patch('eikon.json_requests._check_ticket_async')
    @patch('eikon.Profile.get_profile')
    def test_result_when_post_datagrib_async_with_ticket(self, mock_get_profile, mock_get_ticket):
        mock_profile = Mock_Profile('12345')
        setattr(mock_profile, 'send_request', MagicMock(
            side_effect=[Mock_Response(200, 'x'), Mock_Response(200, {"key1": "value1"})]
        ))
        mock_get_profile.return_value = mock_profile
        mock_get_ticket.side_effect = ['test_ticket', '']

        from eikon.json_requests import send_json_request
        from eikon.data_grid import DataGridAsync_UDF_endpoint
        self.assertEqual("{'key1': 'value1'}", send_json_request(DataGridAsync_UDF_endpoint, {}))

    # @patch('eikon.Profile.get_profile')
    # @patch('eikon.Profile.Profile.get_url')
    # @patch('requests_async.Session.post')
    # def test_debug_output(self, mock_requests_post, mock_get_url, mock_get_profile):
    #     mock_get_profile.return_value = Mock_Profile('12345')
    #     mock_get_url.return_value = 'http://localhost:3000/api/v1/data'
    #     mock_requests_post.return_value = Mock_Response(200, {"key1": "value1"})
    #     from eikon.json_requests import send_json_request
    #     send_json_request('entity', {}, debug=True)
    #     output = sys.stdout.getvalue().strip() # because stdout is an StringIO instance
    #     output = output.split('\n')
    #     first_output = output[0].split(':', 1)
    #     first_output = {first_output[0]: json.loads(first_output[1])}
    #     self.assertEquals(first_output, {'Request': {'Entity': {'E': 'entity', 'W': {}}}})
    #     result = 'HTTP Response: 200 - <bound method Mock_Response'
    #     self.assertEquals(output[1][:len(result)], result)



from unittest import TestCase
from mock import patch
import eikon


class Test_Symbology(TestCase):

    def setUp(self):
        eikon.set_app_key('1234')

    def test_get_symbology_wrong_parameters(self):

        """Check ValueError exception is raised when type of symbol isn't a string nor a list"""
        self.assertRaises(ValueError, eikon.get_symbology, 1234)
        self.assertRaises(ValueError, eikon.get_symbology, {'a': 1})
        self.assertRaises(ValueError, eikon.get_symbology, True)

        """Check ValueeError exception is raised when type of from_symbol isn't a string"""
        self.assertRaises(ValueError, eikon.get_symbology, 'IBM.N', 1234)
        self.assertRaises(ValueError, eikon.get_symbology, 'IBM.N', True)
        self.assertRaises(ValueError, eikon.get_symbology, 'IBM.N', ['123'])
        self.assertRaises(ValueError, eikon.get_symbology, 'IBM.N', {'a': 1})

        """Check ValueError exception is raised when type of to_symbol isn't a string"""
        self.assertRaises(ValueError, eikon.get_symbology, 'IBM.N', 'RIC', 1234)
        self.assertRaises(ValueError, eikon.get_symbology, 'IBM.N', 'RIC', True)
        self.assertRaises(ValueError, eikon.get_symbology, 'IBM.N', 'RIC', [123])
        self.assertRaises(ValueError, eikon.get_symbology, 'IBM.N', 'RIC', {'a': 1})

        """Check ValueError exception is raised when type of to_symbol isn't a list of string"""
        self.assertRaises(ValueError, eikon.get_symbology, 'IBM.N', 'RIC', ['aaa', 1234])
        self.assertRaises(ValueError, eikon.get_symbology, 'IBM.N', 'RIC', ['aaa', True])
        self.assertRaises(ValueError, eikon.get_symbology, 'IBM.N', 'RIC', {'a': 1})

        """Check ValueError exception is raised when from_symbol is unknown symbol type """
        self.assertRaises(ValueError, eikon.get_symbology, 'IBM.N', 'RIQ', 'ISIN')

        """Check ValueError exception is raised when to_symbol contains unknown symbol type """
        self.assertRaises(ValueError, eikon.get_symbology, 'IBM.N', 'RIC', 'WrongSymbolType')
        self.assertRaises(ValueError, eikon.get_symbology, 'IBM.N', ['RIC', 'WrongSymbolType'])

    @patch('eikon.json_requests.send_json_request')
    def test_get_symbology_case_insensitive_symbol_parameters(self, send_json_request):
        json_result_get_symbology_for_IBM = {"mappedSymbols": [
            {'CUSIPs': ['IBM_CUSIP'], 'ISINs': ['IBM_ISIN'], 'RICs': ['IBM.N'], 'SEDOLs': ['IBM_SEDOL'],
             'bestMatch': {'CUSIP': 'IBM_CUSIP', 'ISIN': 'IBM_ISIN', 'RIC': 'IBM.N', 'SEDOL': 'IBM_SEDOL',
                           'ticker': 'IBM'},
             'symbol': 'IBM.N'}]}

        send_json_request.return_value = json_result_get_symbology_for_IBM

        """Check happy path """
        result = eikon.get_symbology('IBM.N', 'RIC', 'ISIN', raw_output=True)
        expected = json_result_get_symbology_for_IBM
        self.assertEqual(result, expected)

        """Check rics ans symbols arn't case sensitive """
        result = eikon.get_symbology('ibm.n', 'Ric', 'Isin', raw_output=True)
        self.assertEqual(result, expected)




# coding: utf-8

__all__ = ['set_app_key', 'get_app_key',
           'set_timeout', 'get_timeout',
           'set_port_number', 'get_port_number',
           'set_log_level', 'set_log_path',
           'set_on_state_callback', 'set_on_event_callback',
           'Profile', 'get_desktop_session']


import deprecation
import logging

from .tools import is_string_type
from .eikonError import EikonError
from eikon import __version__
from .streaming_session import DesktopSession


def set_app_key(app_key):
    """
    Set the app key.

    Parameters
    ----------
    app_key : string
        the app key
    Notes
    -----
    The app key identifies your application on Refinitiv Platform.
    You can get an app key using the App Key Generator (this App is available in Eikon Desktop).
    """
    get_profile().set_app_key(app_key)


def get_app_key():
    """
    Returns the app key previously set

    Notes
    -----
    The app key identifies your application on Refinitiv Platform.
    You can get an application ID using the App Key Generator (this App is available in Eikon Desktop).
    """
    return get_profile().get_app_key()


def get_desktop_session():
    """
    Returns the desktop session for streaming access

    Notes
    -----
    The desktop session is use to initialize streaming subscription.
    """
    return get_profile()._get_desktop_session()


def set_on_state_callback(on_state):
    """
    Set the callback for desktop session state notification.

    Parameters
    ----------
    on_state : function that accepts session, state_ and state_msg as parameters.
    """
    get_profile().set_on_state_callback(on_state)


def set_on_event_callback(on_event):
    """
    Set the callback for desktop session event notification.

    Parameters
    ----------
    on_event : function that accepts session, event_code and event_msg as parameters.
    """
    get_profile().set_on_event_callback(on_event)


def set_timeout(timeout):
    """
    Set the timeout for each request.

    Parameters
    ----------
    timeout : int
        the request timeout in sec
        Default value: 30 sec
    """
    get_profile().set_timeout(timeout)


def get_timeout():
    """
    Returns the request timeout in sec
    """
    return get_profile().get_timeout()


def set_port_number(port_number):
    """
    Set the port number to communicate with the Eikon Data API proxy.
    This port number is detected automatically but you can call this function to force it manually for troubleshooting issues.

    Parameters
    ----------
    port_number : int
        the port number
    """
    get_profile().set_port_number(port_number)


def get_port_number():
    """
    Returns the port number used to communicate with the Eikon Data API Proxy
    """
    return get_profile().get_port_number()


def get_profile():
    """
    Returns the Profile singleton
    """
    return Profile.get_profile()


def set_log_level(level):
    """
    Set the log level.
    When logs are activated (log_level != logging.NOTSET), log files are created in the current directory.
    To change directory for log files, set log path with set_log_path() function.

    Parameters
    ----------
    level : int
        Possible values from logging module : [CRITICAL, FATAL, ERROR, WARNING, WARN, INFO, DEBUG, NOTSET]

    Example
    -------
        ek.set_log_level(logging.DEBUG)

    """
    get_profile().set_log_level(level)


def set_log_path(path):
    """
    Set the filepath of the log file.

    Parameters
    ----------
    path : string
        File path location for log files

    Example
    -------
        ek.set_log_path("c:\\my_directory")
    """
    get_profile().set_log_path(path)


class Profile(object):

    TRACE = 5
    MAX_LOG_SIZE = 10000000

    # singleton profile
    __profile = None

    @classmethod
    def get_profile(cls):
        """
        Returns the Profile singleton
        """
        if cls.__profile is None:
            cls.__profile = Profile()
        return cls.__profile

    def __init__(self):
        """
        Initialization of the __profile.
        """
        self.log_path = None
        self.log_level = logging.NOTSET

        logging.addLevelName(5, 'TRACE')
        self.logger = logging.getLogger('pyeikon')
        setattr(self.logger, 'trace', lambda *args: self.logger.log(5, *args))

        self.app_key = None
        self._desktop_session = None
        self._on_state_cb = None
        self._on_event_cb = None

        self.url = None
        self.streaming_url = None
        self.timeout = None

        self.log_level = logging.NOTSET

    def __del__(self):
        print('Delete a Profile')

    def __delete__(self, instance):
        self.log(1, f'Delete the Profile instance {instance}')

    def set_app_key(self, app_key):
        """
        Set the application id.
        """
        if app_key is None:
            raise AttributeError('App key value is None')
        else:
            if not is_string_type(app_key):
                raise AttributeError('App key must be a string')

            if self._desktop_session:
                self.logger.debug('Reset a Desktop session with new app_key')
                self._desktop_session.close()
                self._desktop_session._app_key = app_key  # = DesktopSession(app_key, self._on_state, self._on_event)
            else:
                self._desktop_session = DesktopSession(app_key=app_key,
                                                       on_state= self._on_state,
                                                       on_event=self._on_event)
            self._desktop_session.set_log_level(self.log_level)
            self._desktop_session.set_timeout(self.timeout)
            self._desktop_session.open()
            self.logger = self._desktop_session.logger()
            self.logger.debug('Init a Desktop session with new app_key')
            self.check_profile()

    def set_on_state_callback(self, on_state):
        self._on_state_cb = on_state

    def set_on_event_callback(self, on_event):
        self._on_event_cb = on_event

    def _on_state(self, session, state_code, state_msg):
        if self._on_state_cb:
            self._on_state_cb(self, state_code, state_msg)

    def _on_event(self, session, event_code, event_msg):
        if self._on_event_cb:
            self._on_event_cb(self, event_code, event_msg)

    def _get_desktop_session(self):
        if self._desktop_session:
            return self._desktop_session
        else:
            raise EikonError(-1, "Set app_key is mandatory first !")

    def get_app_key(self):
        """
        Returns the app key.
        """
        return self._get_desktop_session().app_key

    def get_url(self):
        """
        Returns the scripting proxy url.
        """
        return self._get_desktop_session()._get_udf_url()

    def get_streaming_url(self):
        """
        Returns the streaming_session proxy url.
        """
        return self._get_desktop_session()._streaming_config.url

    def _get_http_session(self):
        """
        Returns the scripting proxy _http_session for requests.
        """
        return self._get_desktop_session()._get_http_session()

    def set_timeout(self, timeout):
        """
        Set the timeout in seconds for each request.
        """
        self.timeout = timeout
        if self._desktop_session:
            self._get_desktop_session().set_timeout(timeout=timeout)

    def get_timeout(self):
        """
        Returns the timeout for requests.
        """
        if self._desktop_session:
            return self._get_desktop_session().get_timeout()
        else:
            return self.timeout

    @deprecation.deprecated(deprecated_in="1.1.3", removed_in="1.2",
                            current_version=__version__,
                            details="Port number is detected automatically.")
    def set_port_number(self, port_number):
        """
        Set the port number to reach Eikon API proxy.
        """
        pass

    @property
    def port(self):
        return self._get_desktop_session().get_port_number()

    def get_port_number(self):
        """
        Returns the port number
        """
        return self._get_desktop_session().get_port_number()

    def set_log_path(self, log_path):
        """
        Set the path where log files will be created.

        Parameters
        ----------
        log_path : path directory
            Default: current directory (beside *.py running file)
        Return True if log_path exists and is writable
        """
        return self._get_desktop_session().set_log_path(log_path)

    def set_log_level(self, log_level):
        """
        Set the log level.
        By default, logs are disabled.

        Parameters
        ----------
        log_level : int
            Possible values from logging module : [CRITICAL, FATAL, ERROR, WARNING, WARN, INFO, DEBUG, NOTSET]
        """
        self.log_level = log_level
        if self._desktop_session:
            self._get_desktop_session().set_log_level(log_level)

    def get_log_level(self):
        """
        Returns the log level
        """
        if self._desktop_session:
            return self._get_desktop_session().get_log_level()
        else:
            return self.log_level

    def check_profile(self):
        if self.port is not None:
            self.logger.info('Port {} on local proxy was detected'.format(self.get_port_number()))
        else:
            # port number wasn't identified => raise EikonError exception
            self.logger.error('Port number was not identified.\nCheck if Eikon Desktop or Eikon API Proxy is running.')
            raise EikonError(-1, 'Port number was not identified. Check if Eikon Desktop or Eikon API Proxy is running.')

    def send_request(self, json, timeout=None):
        response = self._desktop_session.http_request(url=self.get_url(),
                                                      method="POST",
                                                      headers={'Content-Type': 'application/json',
                                                               'x-tr-applicationid': self.get_app_key()},
                                                      json=json,
                                                      timeout=timeout)
        return response



import unittest
# from nose.tools import with_setup, assert_raises
from mock import patch
import logging
import eikon


json_result_get_news_headlines_for_IBM = {'headlines': [{'firstCreated': '2016-01-01T15:04:05',
                                                         'versionCreated': '2016-01-01T15:04:05',
                                                         'storyId': 'storyId',
                                                         'text': 'IBM news headline'}]}

json_result_get_story_for_IBM = {'story': {'headlineHtml': 'headline content',
                                           'storyHtml': 'story content'}}

headline_IBM = ('2016-07-20 15:47:17',{'versionCreated':'2016-07-20 15:47:34',
                                       'text':'AP Top News at 11:44 a.m. EDT<IBM.N><M.N>',
                                       'storyId':'urn:newsml:reuters.com:20160720:nNRA2a4sam:1',
                                       'sourceCode':'NS:ASSOPR','Name':'2016-07-20 15:47:17','dtype':'object'})

result_get_story_for_IBM = 'story content'


class Mock_Profile(object):
    def __init__(self, app_key):
        self.app_key = app_key
        self.port = 3000
        self.url = 'http://localhost:{0}/api/v1/data'.format(self.port)
        self.logger = logging.getLogger('pyeikon')

    def get_app_key(self):
        return self.app_key

    def get_url(self):
        return self.url


class TestNewsHeadlineRequest(unittest.TestCase):

    def setUp(self):
        eikon.set_app_key('1234')

    @patch('eikon.Profile.get_profile')
    def test_raise_AttributeError(self, mock_get_profile ):
        from eikon import get_news_headlines, get_news_story
        mock_get_profile.return_value = Mock_Profile('12345')

        # test get_news_headlines
        # check error on query parameter
        self.assertRaises(ValueError, get_news_headlines, 1234, raw_output=True)
        self.assertRaises(ValueError, get_news_headlines, {'a': 1}, raw_output=True)
        self.assertRaises(ValueError, get_news_headlines, ['a', 1], raw_output=True)

        # check error on headline_count parameter
        self.assertRaises(ValueError, get_news_headlines, 'query', 'abcd', raw_output=True)
        self.assertRaises(ValueError, get_news_headlines, 'query', {'a': 1}, raw_output=True)
        self.assertRaises(ValueError, get_news_headlines, 'query', [1], raw_output=True)
        self.assertRaises(ValueError, get_news_headlines, 'query', -5, raw_output=True)

        # check error on date_from parameter
        self.assertRaises(ValueError, get_news_headlines, 'query', 10, 'abcd', raw_output=True)
        self.assertRaises(ValueError, get_news_headlines, 'query', 10, 1234, raw_output=True)
        self.assertRaises(ValueError, get_news_headlines, 'query', 10, {'a': 1}, raw_output=True)
        self.assertRaises(ValueError, get_news_headlines, 'query', 10, ['a', 1], raw_output=True)

        # check error on date_to parameter
        self.assertRaises(ValueError, get_news_headlines, 'query', 10, '2016-01-01T15:04:05', 'abcd', raw_output=True)
        self.assertRaises(ValueError, get_news_headlines, 'query', 10, '2016-01-01T15:04:05', 1234, raw_output=True)
        self.assertRaises(ValueError, get_news_headlines, 'query', 10, '2016-01-01T15:04:05', {'a': 1}, raw_output=True)
        self.assertRaises(ValueError, get_news_headlines, 'query', 10, '2016-01-01T15:04:05', [1, 2], raw_output=True)

        # test get_news_story
        # check error on story_id parameter
        self.assertRaises(AttributeError, get_news_story, 1234)
        self.assertRaises(AttributeError, get_news_story, {'a': 1})
        self.assertRaises(AttributeError, get_news_story, ['a', 1])

    @patch('eikon.json_requests.send_json_request')
    @patch('eikon.Profile.get_profile')
    def test_get_headlines(self, get_profile, send_json_request):
        from eikon import get_news_headlines
        get_profile.return_value = Mock_Profile('12345')
        send_json_request.return_value = json_result_get_news_headlines_for_IBM
        self.assertEqual(json_result_get_news_headlines_for_IBM, get_news_headlines('IBM', raw_output=True))

    @patch('eikon.json_requests.send_json_request')
    @patch('eikon.Profile.get_profile')
    def test_get_story(self, mock_get_profile, send_json_request):
        from eikon import get_news_story
        mock_get_profile.return_value = Mock_Profile('12345')
        send_json_request.return_value = json_result_get_story_for_IBM
        self.assertEqual(json_result_get_story_for_IBM, get_news_story(headline_IBM, raw_output=True))



# coding: utf-8

__all__ = ['EikonError']


class EikonError(Exception):
    """
    Base class for exceptions specific to Eikon platform.
    """
    def __init__(self, code, message):
        """
        Parameters
        ----------
        code: int

        message: string
            Indicate the sort direction. Possible values are 'asc' or 'desc'. The default value is 'asc'
        """
        self.code = code
        self.message = message

    def __str__(self):
        return 'Error code {} | {}'.format(self.code, self.message)



from unittest import TestCase
from mock import patch, call, Mock, MagicMock
from requests_async import Session
from requests_async.models import Response
from json import dumps

import eikon


payload = {'requests': [{'instruments': ['GOOG.O'], 'fields': [{'name': 'TR.PriceOpen'}, {'name': 'TR.PriceClose'}]}]}


json_result_from_datagrid = {'rowHeadersCount': 1,
                             'data': [['googl.o', 992, 993.27]],
                             'totalRowsCount': 2,
                             'headers': [
                                 [{'displayName': 'Instrument'},
                                  {'displayName': 'Price Open', 'field': 'TR.PRICEOPEN'},
                                  {'displayName': 'Price Close', 'field': 'TR.PRICECLOSE'}]],
                             'columnHeadersCount': 1,
                             'totalColumnsCount': 3,
                             'headerOrientation': 'horizontal'}


class Test_DataGrid(TestCase):

    def setUp(self):
        eikon.set_app_key('1234')

    def test_get_data_wrong_parameters(self):
        ## get_data(instruments, fields, parameters=None, field_name=False, raw_output=False, debug=False)

        # rics must be a string or a list of string
        """Check ValueError exception is raised when type of rics is not string nor list of strings"""
        self.assertRaises(ValueError, eikon.get_data, 'IBM.N', 1234)
        # self.assertRaises(ValueError, eikon.get_data, 'IBM.N', {'a': 1})
        self.assertRaises(ValueError, eikon.get_data, 'IBM.N', True)
        self.assertRaises(ValueError, eikon.get_data, 'IBM.N', [1234])
        #self.assertRaises(ValueError, eikon.get_data, 'IBM.N', [{'a': 1}])
        self.assertRaises(ValueError, eikon.get_data, 'IBM.N', [True])

        """Check ValueError exception is raised when type of fields is not string nor list of strings"""
        self.assertRaises(ValueError, eikon.get_data, 1234, 'TR.Close')
        self.assertRaises(ValueError, eikon.get_data, {'a': 1}, 'TR.Close')
        self.assertRaises(ValueError, eikon.get_data, True, 'TR.Close')
        self.assertRaises(ValueError, eikon.get_data, [1234], 'TR.Close')
        self.assertRaises(ValueError, eikon.get_data, [{'a': 1}], 'TR.Close')
        self.assertRaises(ValueError, eikon.get_data, [True], 'TR.Close')


    """Check if lower case parameters are properly set to upper case"""
    @patch('eikon.json_requests.send_json_request')
    def test_get_data_lower_parameters(self, mock_send_json_request):
        mock_send_json_request.return_value = json_result_from_datagrid

        result = eikon.get_data('goog.o', fields=['high','low','open','close','count','volume'],
                                raw_output=True)
        #expected = [call('DataGrid', payload, False)]


    # check that json_send_request was called with expected parameters
    @patch('eikon.json_requests.send_json_request')
    def test_get_data_result(self, mock_send_json_request):
        mock_send_json_request.return_value = json_result_from_datagrid
        self.assertFalse(mock_send_json_request.called, "called not initialized correctly")
        result = eikon.get_data('goog.o', ['TR.PriceOpen', 'TR.PriceClose'])
        # check that json_send_request was called with expected parameters
        expected = payload
        mock_send_json_request.assert_called_once_with('DataGrid_StandardAsync', expected, debug=False)

    @patch('eikon.json_requests.send_json_request')
    def test_get_data_result_without_headers(self, mock_send_json_request):
        mock_send_json_request.return_value = {'rowHeadersCount': 1,
                                               'data': [['googl.o', 992, 993.27]],
                                               'totalRowsCount': 2,
                                               'columnHeadersCount': 1,
                                               'totalColumnsCount': 3,
                                               'headerOrientation': 'horizontal'}
        self.assertFalse(mock_send_json_request.called, "called not initialized correctly")

        # 1) test field name
        result, err = eikon.get_data('goog.o', ['TR.PriceOpen', 'TR.PriceClose'], field_name=True)
        self.assertIsNone(result)
        self.assertIsNone(err)

        # 2) test without field name
        result, err = eikon.get_data('goog.o', ['TR.PriceOpen', 'TR.PriceClose'], field_name=False)
        self.assertIsNone(result)
        self.assertIsNone(err)

    # check that get_data returns DataFrame with expected headers
    @patch('eikon.json_requests.send_json_request')
    def test_get_data_headers(self, mock_send_json_request):
        mock_send_json_request.return_value = json_result_from_datagrid
        self.assertFalse(mock_send_json_request.called, "called not initialized correctly")

        # test if result contains field names as column headers

        # 1) test field name
        result, err = eikon.get_data('goog.o', ['TR.PriceOpen', 'TR.PriceClose'], field_name=True)
        column_headers = list(result.columns.values)
        self.assertIn('Instrument', column_headers)
        self.assertIn('TR.PRICEOPEN', column_headers)
        self.assertIn('TR.PRICECLOSE', column_headers)
        self.assertNotIn('Price Open', column_headers)

        # 2) test displayt name
        result, err = eikon.get_data('goog.o', ['TR.PriceOpen', 'TR.PriceClose'], field_name=False)
        column_headers = list(result.columns.values)
        self.assertIn('Instrument', column_headers)
        self.assertIn('Price Open', column_headers)
        self.assertIn('Price Close', column_headers)
        self.assertNotIn('TR.PRICEOPEN', column_headers)


    def test_TR_Field(self):
        # rics must be a string or a list of string
        """Check ValueError exception is raised when type of rics is not string nor list of strings"""

        # test wrong values for params
        self.assertRaises(ValueError, eikon.TR_Field, 'Field_Name_1', 1234)
        self.assertRaises(ValueError, eikon.TR_Field, 'Field_Name_1', 'nothing')
        self.assertRaises(ValueError, eikon.TR_Field, 'Field_Name_1', True)
        self.assertRaises(ValueError, eikon.TR_Field, 'Field_Name_1', {})

        # test wrong values for sort_dir
        self.assertRaises(ValueError, eikon.TR_Field, 'Field_Name_1', None, 1234)
        self.assertRaises(ValueError, eikon.TR_Field, 'Field_Name_1', None, 'nothing')
        self.assertRaises(ValueError, eikon.TR_Field, 'Field_Name_1', None, True)
        self.assertRaises(ValueError, eikon.TR_Field, 'Field_Name_1', None, {})

        # test result for sort_dir
        field = eikon.TR_Field('Field_Name_1')
        self.assertEqual(field, {'Field_Name_1':{}})
        field = eikon.TR_Field('Field_Name_1', sort_dir='desc')
        self.assertEqual(field, {'Field_Name_1':{'sort_dir':'desc'}})

        # test wrong values for sort_priority
        self.assertRaises(ValueError, eikon.TR_Field, 'Field_Name_1', None, 'asc', 'nothing')
        self.assertRaises(ValueError, eikon.TR_Field, 'Field_Name_1', None, 'asc', True)
        self.assertRaises(ValueError, eikon.TR_Field, 'Field_Name_1', None, 'asc', {})

        # test result for sort_priority
        field = eikon.TR_Field('Field_Name_1', sort_priority=1)
        self.assertEqual(field, {'Field_Name_1': { 'sort_priority': 1}})

        field = eikon.TR_Field('Field_Name_1', sort_priority=None)
        self.assertEqual(field, {'Field_Name_1': {}})




from lettuce import *
import eikon as eik
from eikon.eikonError import EikonError

__package__ = "lettuce"


@before.each_scenario
def setup_before_each_scenario(scenario):
    """Clean up variables before running each scenarios."""
    world.exception = None
    world.response = None
    eik.set_app_id('')


@step('the daemon is running')
def is_daemon_running(step):
    # TO ADD : mock an http server to simulate that the daemon is running
    pass


@step('the daemon is not running')
def daemon_is_not_running(step):
    pass


@step('application ID is set with a valid EPAID')
def set_a_valid_EPAID(step):
    eik.set_app_id("8A4AD23BBA5182D097D64F")


@step('application ID is set with an invalid EPAID')
def set_an_invalid_EPAID(step):
    eik.set_app_id("INVALID_EPAID")


@step('application ID is not set')
def EPAID_not_set(step):
    eik.set_app_id('')


@step('a JSON request is sent')
def send_a_JSON_request(step):
    payload = '{ "Analysis": [ "OHLCV"], ' \
              '"EndDate": "2015-10-01T10:00:00","StartDate": "2015-09-01T10:00:00", "Tickers": [ "EUR="]}'
    try:
        world.response = eik.send_json_request("TATimeSeries", payload)
    except EikonError as eik_error:
        world.exception = eik_error


@step('a successful response is received')
def successful_response_received(step):
    assert world.exception is None, "An Exception was raised"
    if world.exception is not None:
        assert world.response != '', "Response is empty"


@step('an exception EikonError is raised with text: "(.*)"')
def exception_raised(step, exception_message):
    assert world.exception is not None, "No exception raised"
    assert world.exception.text == exception_message, "Invalid exception message"



# coding: utf8


__all__ = ["StreamingPrice"]


import sys
import logging
import re
from .itemstream import ItemStream
from .stream import StreamState
from .cache import StreamCache
from .streamingprice_callback import StreamingPriceCallback


class StreamingPrice(StreamCache):
    """
    Open a streaming price subscription.

    Parameters
    ----------
    name: string
        RIC to retrieve market prices for.

    service: string
        Specified the service to subscribe on.
        Default: None

    fields: string or list
        Specified the fields to retrieve.
        Default: None

    on_refresh: callable object (streaming_price, message)
        Called when the stream on instrument_name was opened successfully or when the stream is refreshed by the server.
        This callback is called with the reference to the streaming_price object and the instrument full image.
        Default: None

    on_update: callable object (streaming_price, message)
        Called when an update is received for a instrument_name.
        This callback is called with the reference to the streaming_price object and the instrument update.
        Default: None

    on_status: callable object (streaming_price, status)
        Called when a status is received for the stream.
        This callback is called with the reference to the streaming_price object and the instrument status.
        Default: None

    on_complete: callable object  (streaming_price)
        Called when the subscription is completed.
        This callback is called with the reference to the streaming_price object.
        Default: None

    Raises
    ------
    Exception
        If request fails or if Refinitiv Services return an error

    Examples
    --------
    >> import eikon as ek
    >> euro = ek.StreamingPrice('EUR=')
    >> euro.open()
    >> bid = euro.get_field_value('BID')
    >> ask = euro.get_field_value('ASK')
    >>
    >> def on_update(msg):
            ... print(msg)
    >> subscription = ek.StreamingPrice(['VOD.L', 'EUR=', 'PEUP.PA', 'IBM.N'],
            ... ['DSPLY_NAME', 'BID', 'ASK'],
            ... on_update=on_update)
    >> subscription.open()
    {"EUR=":{"DSPLY_NAME":"RBS          LON","BID":1.1221,"ASK":1.1224}}
    {"PEUP.PA":{"DSPLY_NAME":"PEUGEOT","BID":15.145,"ASK":15.155}}
    {"IBM.N":{"DSPLY_NAME":"INTL BUS MACHINE","BID":"","ASK":""}}
    ...
    """

    class Params(object):

        def __init__(self, *args, **kwargs):
            self._name = None
            self._service = None
            self._fields = None
            self._extended_params = None
            self._on_refresh_cb = None
            self._on_update_cb = None
            self._on_status_cb = None
            self._on_complete_cb = None
            self._on_error_cb = None

            self._domain = "MarketPrice"
            self._item_stream = None

            if len(args) > 0 and isinstance(args[0], StreamingPrice.Params):
                self.__init_from_params__(args[0])

            if kwargs:
                self._name = kwargs.get("name")
                self._service = kwargs.get("service")
                self._fields = kwargs.get("fields")
                self._extended_params = kwargs.get("extended_params")
                self._domain = kwargs.get("domain", "MarketPrice")
                self._on_refresh_cb = kwargs.get("on_refresh")
                self._on_status_cb = kwargs.get("on_status")
                self._on_update_cb = kwargs.get("on_update")
                self._on_complete_cb = kwargs.get("on_complete")
                self._on_error_cb = kwargs.get("on_error")

        def __init_from_params__(self, params):
            self._name = getattr(params, "name", None)
            self._service = getattr(params, "service", "IDN_RDF")
            self._fields = getattr(params, "fields", [])
            self._extended_params = getattr(params, "extended_params", None)
            self._on_refresh_cp= getattr(params, "on_refresh", None)
            self._on_status_cb = getattr(params, "on_status", None)
            self._on_update_cb = getattr(params, "on_update", None)
            self._on_complete_cb = getattr(params, "on_complete", None)
            self._on_error_cb = getattr(params, "on_error", None)

        def name(self, name):
            self._name = name
            return self

        def with_fields(self, fields):
            from eikon.tools import build_list
            if fields:
                self._fields = build_list(fields, 'fields')
            else:
                self._fields = None
            return self

        def with_service(self, service):
            if service:
                self._service = service
            return self

        def with_extended_params(self, extended_params):
            if extended_params:
                self._extended_params = extended_params
            return self

        def on_status(self, on_status):
            self._on_status_cb = on_status
            return self

        def on_refresh(self, on_refresh):
            self._on_refresh_cb = on_refresh
            return self

        def on_update(self, on_update):
            self._on_update_cb = on_update
            return self

        def on_complete(self, on_complete):
            self._on_complete_cb = on_complete
            return self

        def on_error(self, on_error):
            self._on_error_cb = on_error
            return self

    def __init__(self,
                 name,
                 session=None,
                 fields=None,
                 service=None,
                 streaming=None,
                 extended_params=None,
                 on_refresh=None,
                 on_status=None,
                 on_update=None,
                 on_complete=None,
                 on_error=None):
        super().__init__(name=name,
                         fields=fields,
                         service=service)

        from ..Profile import get_profile
        if session is None:
            self._session = get_profile()._desktop_session
        else:
            self._session = session
        self._extended_params = extended_params
        self._on_refresh_cb = on_refresh
        self._on_status_cb = on_status
        self._on_update_cb = on_update
        self._on_complete_cb = on_complete
        self._on_error_cb = on_error

        self._callbacks = StreamingPriceCallback()
        self._error_message = None

        if self._session is None:
            raise AttributeError("Session must be defined")
        if self._name is None:
            raise AttributeError("Instrument name must be defined.")

        self._item_stream = ItemStream(session=self._session,
                                       name=self._name,
                                       service=self._service,
                                       fields=self._fields,
                                       on_refresh=self._on_refresh,
                                       on_status=self._on_status,
                                       on_update=self._on_update,
                                       on_complete=self._on_complete,
                                       on_error=self._on_error)


    @property
    def id(self):
        return self._item_stream.stream_id


    def has_error(self):
        return self._error_message

    @property
    def state(self):
        if self._item_stream is None:
            return StreamState.Closed
        else:
            return self._item_stream.state

    @property
    def error_code(self):
        if self._item_stream is None:
            return None
        else:
            return self._item_stream._code

    @property
    def error_message(self):
        if self._item_stream is None:
            return None
        else:
            return self._item_stream._message

    ###################################################
    #  methods to open synchronously item stream      #
    ###################################################
    def open(self, with_updates=True):
        """
        Open the item stream
        """
        self._session.log(logging.DEBUG,
                          f'Open synchronously StreamingSinglePrice {self.id} to {self._name}')
        return self._item_stream.open(with_updates=with_updates)

    def close(self):
        """
        Close the data stream
        """
        self._session.log(logging.DEBUG,
                          f'Stop StreamingSinglePrice subscription {self.id} to {self._name}')
        return self._item_stream.close()

    ################################################
    #  methods to open asynchronously item stream  #
    ################################################
    async def open_async(self, with_updates=True):
        """
        Open the data stream
        """
        self._session.log(logging.DEBUG,
                          f"Open asynchronously StreamingSinglePrice {self.id} to {self._name}")
        await self._item_stream.open_async(with_updates=with_updates)

    ###################################
    # handle decoding partial update   #
    ###################################
    
    #   regular expression pattern for intra-field position sequence
    _partial_update_intra_field_positioning_sequence_regular_expression_pattern = r'[\x1b\x5b|\x9b]([0-9]+)\x60([^\x1b^\x5b|^\x9b]+)'
    _huge = 1e12

    def _decode_intra_field_position_sequence(self, cached_value, new_value):
        ''' this funciton is designed for processing intra-field positioning sequence '''

        #   find all partial update in the value
        tokens = re.findall(self._partial_update_intra_field_positioning_sequence_regular_expression_pattern, new_value)

        #   check this value contains a partial update or not?
        if len(tokens) == 0:
        #   no partial update required, so done
            return new_value
        
        #   do a partial update
        updated_value = cached_value
        for (offset, replace) in tokens:
            #   convert offset from str to int
            offset = int(offset)
            assert(offset<len(updated_value))
            
            #   replace the value in the string
            updated_value = updated_value[:offset] + replace + updated_value[offset+len(replace):]
        
        #   done, return
        return updated_value

    def _decode_partial_update_field(self, key, value):    
        ''' this function is used to proces the partial update 
                RETURNS the processed partial update data
        '''
        
        assert(isinstance(value, str))
        assert('Fields' in self._record)
        assert(key in self._record['Fields'])

        #   prevent attack when using regular expression
        assert(len(value) < self._huge)

        #   process infra-field positioning sequence
        cached_value = self._record['Fields'][key]
        updated_value = self._decode_intra_field_position_sequence(cached_value, value)
        
        #   done
        return updated_value

    ###################################
    # Messages from stream connection #
    ###################################
    def _on_refresh(self, stream, message):
        self._record = message
        if self._on_refresh_cb:
            try:
                self._on_refresh_cb(self, message["Fields"])
            except Exception as e:
                self._session.log(logging.ERROR, f'StreamingPrice on_refresh callback raised exception: {e!r}')
                self._session.log(1, f'Traceback : {sys.exc_info()[2]}')

    def _on_status(self, stream, status):
        self._status = status
        if self._on_status_cb:
            try:
                self._on_status_cb(self, status)
                self._status = status
            except Exception as e:
                self._session.log(logging.ERROR, f'StreamingPrice on_status callback raised exception: {e!r}')
                self._session.log(1, f'Traceback : {sys.exc_info()[2]}')

    def _on_update(self, stream, update):
        for data in update:
            if data == "Fields":
           #   fields data
                #   loop over all update items
                for key, value in update[data].items():
                    #   only string value need to check for a partial update
                    if isinstance(value,str):
                    #   value is a string, so check for partial update string
                        #   process partial update and update the callback with processed partial update
                        update[data][key] = self._decode_partial_update_field( key, value )

                #   update the field data
                self._record[data].update(update[data])
            else:
            #   not a "Fields" data
                self._record[data] = update[data]

        if self._on_update_cb and "Fields" in update:
            try:
                self._on_update_cb(self, update["Fields"])
            except Exception as e:
                self._session.log(logging.ERROR, f'StreamingPrice on_update callback raised exception: {e!r}')
                self._session.log(1, f'Traceback : {sys.exc_info()[2]}')

    def _on_complete(self, stream):
        if self._on_complete_cb:
            try:
                self._on_complete_cb(self)
            except Exception as e:
                self._session.log(logging.ERROR, f'StreamingPrice on_complete callback raised exception: {e!r}')
                self._session.log(1, f'Traceback : {sys.exc_info()[2]}')

    def _on_error(self, stream, error):
        if self._on_error_cb:
            try:
                self._on_error_cb(self, error)
            except Exception as e:
                self._session.log(logging.ERROR, f'StreamingPrice on_error callback raised exception: {e!r}')
                self._session.log(1, f'Traceback : {sys.exc_info()[2]}')



# coding: utf8


__all__ = ["StreamingPrices"]


import sys
import logging
import asyncio

from pandas import DataFrame
from pandas import to_numeric
from .streamingprice import StreamingPrice
from .stream import StreamState


class StreamingPrices:
    """
    Open a streaming price subscription.

    Parameters
    ----------
    instruments: list[string]
        List of RICs to subscribe.

    service: string
        Specified the service to subscribe on.
        Default: None

    fields: string or list[string]
        Specified the fields to retrieve.
        Default: None

    on_refresh: callable object (streaming_prices, instrument_name, message)
        Called when a stream on instrument_name was opened successfully or when the stream is refreshed by the server.
        This callback is called with the reference to the streaming_prices object, the instrument name and the instrument full image.
        Default: None

    on_update: callable object (streaming_prices, instrument_name, message)
        Called when an update is received for a instrument_name.
        This callback is called with the reference to the streaming_prices object, the instrument name and the instrument update.
        Default: None

    on_status: callable object (streaming_prices, instrument_name, status)
        Called when a status is received for a instrument_name.
        This callback is called with the reference to the streaming_prices object, the instrument name and the instrument status.
        Default: None

    on_complete: callable object  (streaming_prices, instrument_name)
        Called when all subscriptions are completed.
        This callback is called with the reference to the streaming_prices object.
        Default: None

    Raises
    ------
    Exception
        If request fails.

    Examples
    --------
    >> import eikon as ek
    >> fx = ek.StreamingPrices(['EUR=', 'GBP='])
    >> fx.open()
    >> bid_eur = fx['EUR']['BID']
    >> ask_eur = fx['EUR']['ASK']
    >>
    >> def on_update(streams, instrument, msg):
            ... print(msg)
    >> subscription = ek.StreamingPrices(['VOD.L', 'EUR=', 'PEUP.PA', 'IBM.N'],
            ... ['DSPLY_NAME', 'BID', 'ASK'],
            ... on_update=on_update)
    >> subscription.open()
    {"EUR=":{"DSPLY_NAME":"RBS          LON","BID":1.1221,"ASK":1.1224}}
    {"PEUP.PA":{"DSPLY_NAME":"PEUGEOT","BID":15.145,"ASK":15.155}}
    {"IBM.N":{"DSPLY_NAME":"INTL BUS MACHINE","BID":"","ASK":""}}
    ...
    """

    class Params(object):
        def __init__(self, instruments, fields):
            self._universe = instruments
            self._fields = fields

        @property
        def instruments(self):
            return self._universe

        @property
        def fields(self):
            return self._fields

    class StreamingPricesIterator:
        """ StreamingPrices Iterator class """
        def __init__(self, streaming_prices):
            self._streaming_prices = streaming_prices
            self._index = 0

        def __next__(self):
            """" Return the next streaming item from streaming price list """
            if self._index < len(self._streaming_prices.params.instruments):
                result = self._streaming_prices[self._streaming_prices.params.instruments[self._index]]
                self._index += 1
                return result
            raise StopIteration()

    def __init__(self,
                 instruments,
                 session=None,
                 fields=[],
                 service=None,
                 on_refresh=None,
                 on_status=None,
                 on_update=None,
                 on_complete=None):
        from eikon.Profile import get_desktop_session
        if session is None:
            self._session = get_desktop_session()
        else:
            self._session = session
        if isinstance(instruments, str):
            instruments = [instruments]
        elif isinstance(instruments, list) and all(isinstance(item, str) for item in instruments):
            pass
        else:
            raise EikonError(-1, "StreamingPrices: instruments must be a list of strings")
        self._fields = fields

        self.params = StreamingPrices.Params(instruments=instruments, fields=fields)

        self._service = service
        self._streaming_prices = {}
        for name in instruments:
            self._streaming_prices[name] = StreamingPrice(session=self._session,
                                                          name=name,
                                                          fields=self._fields,
                                                          service=self._service,
                                                          on_refresh=self._on_refresh,
                                                          on_update=self._on_update,
                                                          on_status=self._on_status,
                                                          on_complete=self._on_complete
                                                          )
        self._on_refresh_cb = on_refresh
        self._on_status_cb = on_status
        self._on_update_cb = on_update
        self._on_complete_cb = on_complete

        self._state = StreamState.Closed

        #   set universe of on_complete 
        self._on_complete_set = None

    @property
    def state(self):
        return self._state

    ###################################################
    #  Access to StreamingPrices as a dict            #
    ###################################################

    def keys(self):
        if self._streaming_prices:
            return self._streaming_prices.keys()
        return {}.keys()

    def values(self):
        if self._streaming_prices:
            return self._streaming_prices.values()
        return {}.values()

    def items(self):
        if self._streaming_prices:
            return self._streaming_prices.items()
        return {}.items()

    ###################################################
    #  Make StreamingPrices iterable                  #
    ###################################################

    def __iter__(self):
        return StreamingPrices.StreamingPricesIterator(self)

    def __getitem__(self, item):
        if item in self.params.instruments:
            return self._streaming_prices[item]
        else:
            raise KeyError(f"{item} not in StreamingPrices universe")

    def __len__(self):
        return len(self.params.instruments)

    ###################################################
    #  methods to open synchronously item stream      #
    ###################################################
    def open(self, with_updates=True):
        """
        Open synchronously the streaming price
        """
        return self._session._loop.run_until_complete(self.open_async(with_updates=with_updates))

    ################################################
    #  methods to open asynchronously item stream  #
    ################################################
    async def open_async(self, with_updates=True):
        """
        Open asynchronously the streaming price
        """
        self._session.log(1, f'StreamingPrices : open streaming on {self.params.instruments}')
        if self._state == StreamState.Open:
            return

        self._state = StreamState.Pending
        self._on_complete_set = set()

        task_list = [stream.open_async(with_updates=with_updates) for stream in self._streaming_prices.values()]
        await asyncio.wait(task_list, return_when=asyncio.ALL_COMPLETED)
        self._state = StreamState.Open
        self._session.log(1, f'StreamingPrices : start asynchrously streaming on {self.params.instruments} done')
        return self._state

    def close(self):
        if self._state is not StreamState.Closed:
            self._session.log(1, f'StreamingPrices : close streaming on {self.params.instruments}')
            for stream in self._streaming_prices.values():
                stream.close()
        self._state = StreamState.Closed
        return self._state

    def get_snapshot(self, instruments=None, fields=None, convert=True):
        """
        Returns a Dataframe filled with snapshot values for a list of instrument names and a list of fields.

        Parameters
        ----------
        instruments: list of strings
            List of instruments to request snapshot data on.

        fields: list of strings
            List of fields to request.

        convert: boolean
            If True, force numeric conversion for all values.

        Returns
        -------
            pandas.DataFrame

            pandas.DataFrame content:
                - columns : instrument and fieled names
                - rows : instrument name and field values

        Raises
        ------
            Exception
                If request fails or if server returns an error

            ValueError
                If a parameter type or value is wrong

        Examples
        --------
        >>> import eikon as ek
        >>> ek.set_app_key('set your app key here')
        >>> streaming_prices = ek.StreamingPrices(instruments=["MSFT.O", "GOOG.O", "IBM.N"], fields=["BID", "ASK", "OPEN_PRC"])
        >>> data = streaming_prices.get_snapshot(["MSFT.O", "GOOG.O"], ["BID", "ASK"])
        >>> data
              Instrument    BID        ASK
        0     MSFT.O        150.9000   150.9500
        1     GOOG.O        1323.9000  1327.7900
        2     IBM.N         NaN        NaN
        """
        from eikon.eikonError import EikonError

        if instruments:
            for name in instruments:
                if name not in self.params.instruments:
                    raise ElektronError(-1, f'Instrument {name} was not requested : {self.params.instruments}')

        if fields:
            for field in fields:
                if field not in self.params.fields:
                    raise EikonError(-1, f'Field {field} was not requested : {self.params.fields}')

        _universe = instruments if instruments else self.params.instruments
        _all_fields_value = {name: self._streaming_prices[name].get_fields(fields)
        if name in self._streaming_prices else None
                             for name in _universe}
        _fields = []

        if not fields:
            fields = []
            for field_values in _all_fields_value.values():
                if field_values:
                    _fields.extend(field for field in field_values.keys() if field not in _fields)
        else:
            _fields = fields

        _df_source = {f: [_all_fields_value[name][f] if _all_fields_value[name].get(f) else None
                          for name in _universe] for f in _fields}
        _price_dataframe = DataFrame(_df_source, columns=_fields)
        if convert:
            _price_dataframe = _price_dataframe.apply(to_numeric, errors='ignore')
        _price_dataframe.insert(0, 'Instrument', _universe)
        if convert and _df_source:
            _price_dataframe = _price_dataframe.convert_dtypes()

        return _price_dataframe

    #########################################
    # Messages from stream_cache connection #
    #########################################
    def _on_refresh(self, stream, message):
        if self._on_refresh_cb:
            try:
                self._session.log(1, 'StreamingPrices : call on_refresh callback')
                self._state = StreamState.Open
                self._session._loop.call_soon_threadsafe(self._on_refresh_cb, self, stream.name, message)
                # self._on_refresh_cb(self, name, message)
            except Exception as e:
                self._session.log(logging.ERROR, f'StreamingPrices on_refresh callback raised exception: {e!r}')
                self._session.log(1, f'Traceback : {sys.exc_info()[2]}')

    def _on_status(self, stream, status):

        if self._on_status_cb:
            try:
                self._session.log(1, 'StreamingPrices : call on_status callback')
                self._session._loop.call_soon_threadsafe(self._on_status_cb, self, stream.name, status)
            except Exception as e:
                self._session.log(logging.ERROR, f'StreamingPrices on_status callback raised exception: {e!r}')
                self._session.log(1, f'Traceback : {sys.exc_info()[2]}')

        #   check for closed stream when status "Closed", "ClosedRecover", "NonStreaming" or "Redirect"
        if stream.state == StreamState.Closed and stream.name not in self._on_complete_set:
            #   this stream has been closed, so it means completed also
            self._on_complete(stream)

    def _on_update(self, stream, update):
        if self._on_update_cb:
            try:
                self._session.log(1, 'StreamingPrices : call on_update callback')
                self._session._loop.call_soon_threadsafe(self._on_update_cb, self, stream.name, update)
            except Exception as e:
                self._session.log(logging.ERROR, f'StreamingPrices on_update callback raised exception: {e!r}')
                self._session.log(1, f'Traceback : {sys.exc_info()[2]}')

    def _on_complete(self, stream):
        assert self._on_complete_set is not None

        #   check for update completed set
        if stream.name not in self._on_complete_set:
            #   update the stream to be in complete list
            self._on_complete_set.update([stream.name, ])

            #   check for complete for all subscribe universe
            if self._on_complete_set == set(self.params.instruments):
                if self._on_complete_cb:
                    try:
                        self._session.log(1, 'StreamingPrices : call on_complete callback')
                        self._session._loop.call_soon_threadsafe(self._on_complete_cb, self)
                    except Exception as e:
                        self._session.log(logging.ERROR, f'StreamingPrices on_complete callback raised exception: {e!r}')
                        self._session.log(1, f'Traceback : {sys.exc_info()[2]}')


# coding: utf-8
__version__ = '1.1.10'

"""
    Eikon Data API for Python allows your Python applications to access Refinitiv data directly from Eikon.
    Its usage requires:
        - An App Key (you can create it wit App Key Generator in Eikon Desktop)
        - Eikon Desktop or Refinitiv Workspace application running on your local machine

"""

from .Profile import *
from .symbology import get_symbology
from .json_requests import send_json_request
from .news_request import get_news_headlines, get_news_story
from .time_series import get_timeseries
from .data_grid import get_data, TR_Field
from .eikonError import EikonError
from .streaming_session import *



from unittest import TestCase
from mock import patch, call, Mock, MagicMock
from requests_async import Session
from requests_async.models import Response
from json import dumps

import eikon


json_result_from_timeseries = {'timeseriesData': [
    {'dataPoints': [
        ['2016-12-01T00:00:00Z', 194.207808, 190.375, 192.2, 191, 19059, 104508409],
        ['2016-12-02T00:00:00Z', 192.7163, 188.85, 191.1, 191.4, 15866, 92887431]],
        'fields': [
            {'name': 'TIMESTAMP', 'type': 'DateTime'}, {'name': 'HIGH', 'type': 'Double'},
            {'name': 'LOW', 'type': 'Double'}, {'name': 'OPEN', 'type': 'Double'},
            {'name': 'CLOSE', 'type': 'Double'}, {'name': 'COUNT', 'type': 'Double'},
            {'name': 'VOLUME', 'type': 'Double'}],
        'ric': 'VOD.L',
        'statusCode': 'Normal'
    }
]}


class Test_Time_Series(TestCase):

    def setUp(self):
        eikon.set_app_key('1234')

    def test_get_timeseries_wrong_parameters(self):

        # rics must be a string or a list of string
        """Check ValueError exception is raised when type of rics is not string nor list of strings"""
        self.assertRaises(ValueError, eikon.get_timeseries, 1234, None, '2016-01-01T15:04:05', '2016-02-01T15:04:05')
        self.assertRaises(ValueError, eikon.get_timeseries, {'a': 1}, None, '2016-01-01T15:04:05', '2016-02-01T15:04:05')
        self.assertRaises(ValueError, eikon.get_timeseries, True, None, '2016-01-01T15:04:05', '2016-02-01T15:04:05')
        self.assertRaises(ValueError, eikon.get_timeseries, [1234], None, '2016-01-01T15:04:05', '2016-02-01T15:04:05')
        self.assertRaises(ValueError, eikon.get_timeseries, [{'a': 1}], None, '2016-01-01T15:04:05', '2016-02-01T15:04:05')
        self.assertRaises(ValueError, eikon.get_timeseries, [True], None, '2016-01-01T15:04:05', '2016-02-01T15:04:05')

        """Check ValueError exception is raised when start_date isn't a datetime"""
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, 'abcd', '2016-02-01T15:04:05')
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, 1234, '2016-02-01T15:04:05')
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, {'a': 1}, '2016-02-01T15:04:05')
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, ['a', 1], '2016-02-01T15:04:05')

        """Check ValueError exception is raised when end_date isn't a datetime"""
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', 'abcd')
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', 1234)
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', {'a': 1})
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', [1, 2])

        """Check ValueError exception is raised when type of interval is not string"""
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', 1234)
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', True)
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', ['123'])
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', {'a': 1})

        """Check ValueError exception is raised when fields is not in type list [None, string, list]"""
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', 1234)
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', True)
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', [123])
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', {'a': 1})

        """Check ValueError exception is raised when count is not an integer"""
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', None, True)
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', None, [123])
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', None, {'a': 1})

        """Check ValueError exception is raised when type of calendar is not string"""
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', None, None, 1234)
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', None, None, True)
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', None, None, ['123'])
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', None, None, {'a': 1})

        """Check ValueError exception is raised when type of calendar is not string"""
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', None, None, 1234)
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', None, None, True)
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', None, None, ['123'])
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', None, None, {'a': 1})

        """Check ValueError exception is raised when type of corax is not string"""
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', None, None, None, 1234)
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', None, None, None, True)
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', None, None, None, ['123'])
        self.assertRaises(ValueError, eikon.get_timeseries, 'IBM.N', None, '2016-01-01T15:04:05', '2016-02-01T15:04:05', None, None, None, {'a': 1})


    """Check if lower case parameters are properly set to upper case"""
    @patch('eikon.json_requests.send_json_request')
    def test_get_timeseries_lower_parameters(self, mock_send_json_request):
        mock_send_json_request.return_value = json_result_from_timeseries

        self.assertFalse(mock_send_json_request.called, "called not initialized correctly")

        payload = {'rics': ['GOOG.O'],
                   'interval': 'daily',
                   'enddate': '2016-01-05T15:04:05-07:00',
                   'startdate': '2016-01-01T15:04:05-07:00',
                   'fields': ['HIGH', 'LOW', 'OPEN', 'CLOSE', 'COUNT', 'VOLUME', 'TIMESTAMP']}

        result = eikon.get_timeseries('goog.o', fields=['high','low','open','close','count','volume'],
                                start_date='2016-01-01T15:04:05-07:00', end_date='2016-01-05T15:04:05-07:00',
                                interval='daily')

        # check that json_send_request was called with expected parameters
        mock_send_json_request.assert_called_once_with('TimeSeries', payload, debug=False)




# coding: utf-8

from unittest import TestCase
from requests_async import HTTPError
from datetime import datetime, timedelta
from dateutil.tz import tzlocal
from time import strftime


class TestTools(TestCase):
    def test_isStringType(self):
        from eikon.tools import is_string_type
        self.assertTrue(is_string_type('string'))
        self.assertTrue(is_string_type(u'string'))
        self.assertFalse(is_string_type(1234))
        self.assertFalse(is_string_type(['string']))
        self.assertFalse(is_string_type({'string'}))

    def test_check_for_string(self):
        from eikon.tools import check_for_string
        self.assertIsNone(check_for_string('string', 'test'))
        self.assertIsNone(check_for_string(u'string', 'test'))
        self.assertRaises(ValueError, check_for_string, 1234, 'test')
        self.assertRaises(ValueError, check_for_string, ['string'], 'test')
        self.assertRaises(ValueError, check_for_string, {'key':'value'}, 'test')

    def test_check_for_string_or_list_of_strings(self):
        from eikon.tools import check_for_string_or_list_of_strings
        self.assertIsNone(check_for_string_or_list_of_strings('string', 'test'))
        self.assertIsNone(check_for_string_or_list_of_strings(u'string', 'test'))
        self.assertIsNone(check_for_string_or_list_of_strings(['string1', 'string2'], 'test'))
        self.assertIsNone(check_for_string_or_list_of_strings(['string'], 'test'))
        self.assertRaises(ValueError, check_for_string_or_list_of_strings, 1234, 'test')
        self.assertRaises(ValueError, check_for_string_or_list_of_strings, ['string1', 123], 'test')
        self.assertRaises(ValueError, check_for_string_or_list_of_strings, {'string'}, 'test')

    def test_get_json_value(self):
        from eikon.tools import get_json_value
        self.assertEqual('value', get_json_value({'name': 'value'}, 'name'))
        self.assertEqual(None, get_json_value({'name2': 'value'}, 'name1'))

    def test_check_server_error(self):
        from eikon.json_requests import _check_server_error
        from eikon import EikonError
        class Response():
            ErrorCode='403'
            ErrorMessage='Forbidden'
        self.assertRaises(HTTPError, _check_server_error, Response())
        self.assertRaises(HTTPError, _check_server_error, '<HTTP error>')
        self.assertRaises(EikonError, _check_server_error, {'ErrorCode': 500, 'ErrorMessage': 'error message'})
        self.assertIsNone(_check_server_error('normal response'))

    def test_to_datetime(self):
        from eikon.tools import to_datetime
        self.assertEqual(to_datetime('2016-01-01T15:04:05'), datetime(2016, 1, 1, 15, 4, 5))
        self.assertEqual(to_datetime('2016-01-01 15:04:05'), datetime(2016, 1, 1, 15, 4, 5))
        self.assertEqual(to_datetime('2016-01-01'), datetime(2016, 1, 1))
        self.assertRaises(ValueError, to_datetime, 'everything but a date')

        # check date and time are correctly computed
        #  (ignore milliseconds with timetuple() function)
        dt1 = datetime.now(tzlocal()) + timedelta(-10)
        dt2 = to_datetime(timedelta(-10))
        tt1 = dt1.timetuple()
        tt2 = dt2.timetuple()
        self.assertEqual(tt1,tt2)

    def test_get_date_from_today(self):
        from eikon.tools import get_date_from_today, to_datetime
        self.assertRaises(ValueError, to_datetime, 'everything but a date')

        date_value = datetime.date(to_datetime(timedelta(-10)))
        self.assertEqual(datetime.date(get_date_from_today(10)), date_value)



from unittest import TestCase
from mock import patch
from nose.tools import raises
import os, glob, logging


class TestProfile(TestCase):

    # Remove *.log files before each test
    @classmethod
    def setup_class(cls):
        for filename in os.listdir(os.curdir):
            if filename.endswith('.log'):
                os.remove(filename)

    @raises(AttributeError)
    def test_profile_with_wrong_appid_1(self):
        from eikon.Profile import get_profile, set_app_key
        get_profile().set_app_key([])

    @raises(AttributeError)
    def test_profile_with_wrong_appid_2(self):
        from eikon.Profile import get_profile, set_app_key
        get_profile().set_app_key({'a': 'b'})

    @raises(AttributeError)
    def test_profile_with_wrong_appid_3(self):
        from eikon.Profile import get_profile, set_app_key
        get_profile().set_app_key(12345)

    @patch('os.path.exists')
    @patch('eikon.Profile.read_firstline_in_file')
    @patch('eikon.Profile.check_port')
    def test_get_scripting_proxy_port(self, mock_check_port, read_firstline_in_file, mock_os_path_exits):
        from eikon.Profile import identify_scripting_proxy_port
        mock_os_path_exits.return_value = True

        from requests_async import Session
        session = Session()

        # test when .portInUse file is empty or doesn't exist
        read_firstline_in_file.return_value = ''
        mock_check_port.return_value = False
        self.assertEqual(None, identify_scripting_proxy_port(session, 'app_key'))

        # test when .portInUse file exists, contains validated '123' port number
        read_firstline_in_file.return_value = '123'
        mock_check_port.return_value = True
        self.assertEqual('123', identify_scripting_proxy_port(session, 'app_key'))

        # test when .portInUse file exists, contains '123' port number, but Proxy is listening on 9000 port
        read_firstline_in_file.return_value = '123'
        def fake_check_port(session, application_key, port, timeout=1.0):
            return True if port == '9000' else False
        mock_check_port.side_effect = fake_check_port
        result = identify_scripting_proxy_port(session, 'app_key')
        self.assertEqual('9000', result)

        # test when .portInUse file is empty or doesn't exist and Proxy is listening on 36036 port
        read_firstline_in_file.return_value = ''
        def fake_check_port(session, application_key, port, timeout=1.0):
            return True if port == '36036' else False
        mock_check_port.side_effect = fake_check_port
        self.assertEqual('36036', identify_scripting_proxy_port(session, 'app_key'))

    def test_log_path(self):
        import logging
        import glob

        # clean old pyeikon.*.log files in TEMP directory
        file_pattern = os.path.join(os.environ.get('TEMP'), 'pyeikon*.log')
        for filename in glob.glob(file_pattern):
            os.remove(filename)

        # check no pyeikon.*.log file is in TEMP directory
        self.assertFalse(glob.glob(file_pattern))
        from eikon.Profile import Profile
        profile = Profile()
        profile.set_log_path(os.environ.get('TEMP'))
        profile.set_log_level(logging.INFO)
        profile.set_app_key('123')

        # check that a pyeikon.*.log was created
        self.assertTrue(glob.glob(file_pattern))

    def test_logger(self):
        import logging

        # clean old pyeikon.*.log files in current directory
        file_pattern = os.path.join(os.curdir, 'pyeikon*.log')
        for filename in glob.glob(file_pattern):
            os.remove(filename)

        # check that no pyeikon.*.log is in current directory
        filenames = os.listdir(os.curdir)
        self.assertFalse(any((filename.endswith('.log')) for filename in filenames))

        from eikon.Profile import Profile
        profile = Profile()
        profile.set_log_path(os.curdir)
        profile.set_log_level(logging.INFO)
        profile.set_app_key('123')

        # check that a pyeikon.*.log was created
        filenames = os.listdir(os.curdir)
        self.assertTrue(any((filename.endswith('.log')) for filename in filenames))



# coding: utf-8

import dateutil.parser
import json
from datetime import date, datetime, timedelta
from dateutil.tz import tzlocal


def is_string_type(value):
    try:
        return isinstance(value, basestring)
    except NameError:
        return isinstance(value, str)


def get_json_value(json_data, name):
    if name in json_data:
        return json_data[name]
    else:
        return None


def to_datetime(date_value):
    if type(date_value) is timedelta:
        return datetime.now(tzlocal()) + date_value

    if type(date_value) in (datetime, date):
        return date_value

    try:
        return dateutil.parser.parse(date_value)
    except ValueError as e:
        raise e
    except Exception as e:
        raise ValueError(e)


def get_date_from_today(days_count): 
    if type(days_count) != int :
        raise ValueError('The parameter {} should be an integer, found {}'.format(days_count,type(days_count)))
    return datetime.now(tzlocal()) + timedelta(days=-days_count)


def is_list_of_string(values):
  return all(is_string_type(value) for value in values)


def check_for_string(parameter, name):
    if not is_string_type(parameter):
        raise ValueError('The parameter {} should be a string, found {}'.format(name,str(parameter)))


def check_for_string_or_list_of_strings(parameter, name):
    if type(parameter) != list and (not parameter or not is_string_type(parameter)):
        raise ValueError('The parameter {} should be a string or a list of string, found {}'.format(name,type(parameter)))
    if type(parameter) == list and not is_list_of_string(parameter): 
        raise ValueError('All items in the parameter {} should be of data type string, found {}'.format(name,[type(v) for v in parameter]))


def check_for_int(parameter, name):
    if type(parameter) is not int:
        raise ValueError('The parameter {} should be an int, found {} type value ({})'.format(name, type(parameter), str(parameter)))


def build_list_with_params(values, name):
    if values is None:
        raise ValueError(name + ' is None, it must be a string or a list of strings')

    if is_string_type(values):
        return [(v, None) for v in values.split()]
    elif type(values) is list:
        try:
            return [(value, None) if is_string_type(value) else (value[0], value[1]) for value in values]
        except Exception:
            raise ValueError(name + ' must be a string or a list of strings or a tuple or a list of tuple')
    else:
        try:
            return values[0], values[1]
        except Exception:
            raise ValueError(name + ' must be a string or a list of strings or a tuple or a list of tuple')


def build_list(values, name):
    if values is None:
        raise ValueError(name + ' is None, it must be a string or a list of strings')

    if is_string_type(values):
        return [values.strip()]
    elif type(values) is list:
        if all(is_string_type(value) for value in values):
            return [value for value in values]
        else:
            raise ValueError(name + ' must be a string or a list of strings')
    else:
        raise ValueError(name + ' must be a string or a list of strings')


def build_dictionary(dic, name):
    if dic is None:
        raise ValueError(name + ' is None, it must be a string or a dictionary of strings')

    if is_string_type(dic):
        return json.loads(dic)
    elif type(dic) is dict:
        return dic
    else:
        raise ValueError(name + ' must be a string or a dictionary')


def tz_replacer(s):
    if isinstance(s, str):
        if s.endswith('Z'):
            s = s[:-1]
        elif s.endswith('-0000'):
            s = s[:-5]
        if s.endswith('.000'):
            s = s[:-4]
    return s


# coding: utf8

__all__ = ["StreamCache"]

import copy


class StreamCache:
    """
    StreamCache contains all retrieved data from stream subscription.

    Raises
    ------
    Exception
        If request fails or if Refinitiv Services return an error
    """

    class StreamCacheIterator:
        """ StreamCache Iterator class """

        def __init__(self, stream_cache):
            if stream_cache.get_fields():
                self._field_values = list(stream_cache.get_fields().items())
            else:
                self._field_values = []
            self._index = 0

        def __next__(self):
            """" Return the next field value from stream cache """
            if self._index < len(self._field_values):
                result = self._field_values[self._index]
                self._index += 1
                return result
            raise StopIteration()

    def __init__(self,
                 name,
                 fields=None,
                 service=None,
                 status=None,
                 record=None):
        self._name = name
        self._fields = fields if fields else []
        self._service = service
        self._status = status
        self._record = record

    ###################################################
    #  Access to StreamCache as a dict                #
    ###################################################

    def keys(self):
        if self._record and self._record.get("Fields"):
            return self._record["Fields"].keys()
        return {}.keys()

    def values(self):
        if self._record and self._record.get("Fields"):
            return self._record["Fields"].values()
        return {}.values()

    def items(self):
        if self._record and self._record.get("Fields"):
            return self._record["Fields"].items()
        return {}.items()

    ###################################################
    #  Make StreamCache iterable                      #
    ###################################################

    def __iter__(self):
        return StreamCache.StreamCacheIterator(self)

    def __getitem__(self, field):
        if self._record and self._record.get("Fields"):
            if field in self._record["Fields"].keys():
                return self._record["Fields"][field]
        raise KeyError(f"Field '{field}' not in Stream cache")

    def __len__(self):
        return len(self._fields)

    ###################################################
    #  StreamCache properties                         #
    ###################################################

    @property
    def name(self):
        return self._name

    @property
    def service(self):
        return self._service

    @property
    def fields(self):
        if self._record and self._record.get("Fields"):
            return self._record["Fields"].keys()
        return None

    @property
    def status(self):
        return self._status

    @property
    def is_ok(self):
        return True if self._status.get("Data") == "Ok" else False

    # ###################################################
    # #  StreamingCache data accessors                  #
    # ###################################################

    def get_field_value(self, field):
        if self._record and self._record.get("Fields"):
            if field in self._record["Fields"].keys():
                return self._record["Fields"][field]
        #return None

    def get_fields(self, fields=None):
        if self._record:
            if fields:
                all_fields = self._record.get("Fields")
                selected_fields = {}
                for f in fields:
                    if f in all_fields:
                        selected_fields[f] = all_fields[f]
                    else:
                        selected_fields[f] = None
                return selected_fields
            else:
                return self._record.get("Fields")
        else:
            if fields:
                return {field: None for field in fields}
            else:
                return {field: None for field in self._fields}
